// Emacs style mode select	 -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:
//		DOOM Network game communication and protocol,
//		all OS independent parts.
//
//-----------------------------------------------------------------------------

#include <stddef.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>

#include "dstrings.h"

#include "version.h"
#include "m_alloc.h"
#include "m_menu.h"
#include "m_random.h"
#include "i_system.h"
#include "i_video.h"
#include "i_net.h"
#include "g_game.h"
#include "doomdef.h"
#include "doomstat.h"
#include "c_console.h"
#include "d_netinf.h"
#include "cmdlib.h"
#include "s_sound.h"
#include "m_cheat.h"
#include "p_effect.h"
#include "p_local.h"
#include "c_dispatch.h"
#include "cl_main.h"
#include "sv_main.h"

#include "gi.h"

#define NCMD_EXIT				0x80000000
#define NCMD_RETRANSMIT 		0x40000000
#define NCMD_SETUP				0x20000000
#define NCMD_KILL				0x10000000		// kill game
#define NCMD_CHECKSUM			0x0fffffff

extern byte		*demo_p;		// [RH] Special "ticcmds" get recorded in demos

doomcom_t*		doomcom;		
doomdata_t* 	netbuffer;		// points inside doomcom
netadr_t	localaddr;



//
// NETWORKING
//
// gametic is the tic about to (or currently being) run
// maketic is the tick that hasn't had control made for it yet
// nettics[] has the maketics for all players 
//
// a gametic cannot be run until nettics[] > gametic for all players
//
#define RESENDCOUNT 	10
#define PL_DRONE		0x80	// bit flag in doomdata->player


FDynamicBuffer	NetSpecs[MAXPLAYERS][BACKUPTICS];
ticcmd_t		netcmds[MAXPLAYERS][BACKUPTICS];
int 			nettics[MAXNETNODES];
BOOL 			nodeingame[MAXNETNODES];				// set false as nodes leave game
BOOL	 		remoteresend[MAXNETNODES];				// set when local needs tics
int 			resendto[MAXNETNODES];					// set when remote needs tics
int 			resendcount[MAXNETNODES];

unsigned int	lastrecvtime[MAXPLAYERS];				// [RH] Used for pings
unsigned int	currrecvtime[MAXPLAYERS];

int 			nodeforplayer[MAXPLAYERS];

int 			maketic;
int 			lastnettic;
int 			skiptics;
int 			maxsend;

void D_ProcessEvents (void); 
void G_BuildTiccmd (ticcmd_t *cmd); 
void D_DoAdvanceDemo (void);
 
BOOL	 		reboundpacket;
doomdata_t		reboundstore;

// [RH] Special "ticcmds" get stored in here
static struct TicSpecial
{
	byte *streams[BACKUPTICS];
	int	  used[BACKUPTICS];
	byte *streamptr;
	int	  streamoffs;
	int   specialsize;
	int	  lastmaketic;
	BOOL  okay;

	TicSpecial ()
	{
		int i;

		lastmaketic = -1;
		specialsize = 256;

		for (i = 0; i < BACKUPTICS; i++)
			streams[i] = NULL;

		for (i = 0; i < BACKUPTICS; i++)
		{
			streams[i] = (byte *)Malloc (256);
			used[i] = 0;
		}
		okay = true;
	}

	~TicSpecial ()
	{
		int i;

		for (i = 0; i < BACKUPTICS; i++)
		{
			if (streams[i])
			{
				free (streams[i]);
				streams[i] = NULL;
				used[i] = 0;
			}
		}
		okay = false;
	}

	// Make more room for special commands.
	void GetMoreSpace ()
	{
		int i;

		specialsize <<= 1;

		DPrintf ("Expanding special size to %d\n", specialsize);

		for (i = 0; i < BACKUPTICS; i++)
			streams[i] = (byte *)Realloc (streams[i], specialsize);

		streamptr = streams[maketic%BACKUPTICS] + streamoffs;
	}

	void CheckSpace (int needed)
	{
		if (streamoffs >= specialsize - needed)
			GetMoreSpace ();

		streamoffs += needed;
	}

	void NewMakeTic ()
	{
		if (lastmaketic != -1)
			used[lastmaketic%BACKUPTICS] = streamoffs;

		lastmaketic = maketic;
		streamptr = streams[maketic%BACKUPTICS];
		streamoffs = 0;
	}

	TicSpecial &operator << (byte it)
	{
		if (streamptr)
		{
			CheckSpace (1);
			WriteByte (it, &streamptr);
		}
		return *this;
	}

	TicSpecial &operator << (short it)
	{
		if (streamptr)
		{
			CheckSpace (2);
			WriteWord (it, &streamptr);
		}
		return *this;
	}

	TicSpecial &operator << (int it)
	{
		if (streamptr)
		{
			CheckSpace (4);
			WriteLong (it, &streamptr);
		}
		return *this;
	}

	TicSpecial &operator << (const char *it)
	{
		if (streamptr)
		{
			CheckSpace (strlen (it) + 1);
			WriteString (it, &streamptr);
		}
		return *this;
	}

} specials;

//
// NetUpdate
// Builds ticcmds for console player,
// sends out a packet
//

int nettime;

void NetUpdate (void)
{
    int     nowtime;
    int     newtics;
    int     i;

	if (netstate == NS_SERVER)
		return;

    // check time
    nowtime = I_GetTime ();
    newtics = nowtime - nettime;
    nettime = nowtime;

    if (newtics <= 0)       // nothing new to update
       return;

    for (i = 0; i < newtics; i++)
    {
        I_StartTic ();
        D_ProcessEvents ();
        G_BuildTiccmd (&netcmds[consoleplayer][maketic%BACKUPTICS]);
        maketic++;
    }
}



//
// CheckAbort
//
BOOL CheckAbort (void)
{
	event_t *ev;

	Printf (PRINT_HIGH, "");	// [RH] Give the console a chance to redraw itself
	// This WaitForTic is to avoid flooding the network with packets on startup.
	I_WaitForTic (I_GetTime () + TICRATE/4);
	I_StartTic ();
	for ( ; eventtail != eventhead 
		  ; eventtail = (++eventtail)&(MAXEVENTS-1) ) 
	{ 
		ev = &events[eventtail]; 
		if (ev->type == ev_keydown && ev->data1 == KEY_ESCAPE)
			return true;
	}
	return false;
}


//
// D_CheckNetGame
// Works out player numbers among the net participants
//
extern int viewangleoffset;

void D_CheckNetGame (void)
{
	if (Args.CheckParm("-connect"))
		CL_InitNetwork ();
	else if (Args.CheckParm("-host"))
		SV_InitNetwork ();
	else
		netstate = NS_SINGLEPLAYER;

	if (netstate == NS_SERVER)
	{
		consoleplayer = displayplayer = 0;

		for (int i=0; i<MAXPLAYERS; i++)
			playeringame[i] = false;

		return;
	}

    netbuffer = &doomcom->data;
//	consoleplayer = displayplayer = doomcom->consoleplayer;
	consoleplayer = displayplayer = 0;

    // read values out of doomcom
    maxsend = BACKUPTICS/2 - 1;
    if (maxsend<1)
        maxsend = 1;

    displayplayer = consoleplayer = 0;

	// [BC] Get our local IP
	localaddr = NET_GetLocalAddress ();

	playeringame[consoleplayer] = true;
}


//
// D_QuitNetGame
// Called before quitting to leave a net game
// without hanging the other players
//
void STACK_ARGS D_QuitNetGame (void)
{
	if (debugfile)
		fclose (debugfile);

	if (!usergame || consoleplayer == -1 || demoplayback || netstate == NS_SERVER)
	{
		return;
	}

	if (netstate == NS_CLIENT)
	{
		CL_QuitNetGame ();
	}
}

//
// TryRunTics
//
int 	frameon;
int 	frameskip[MAXPLAYERS];
int 	oldnettics;

extern	BOOL	 advancedemo;

void TryRunTics (void)
{
    int             entertic;
    static int      oldentertics;
    int             realtics;

    // get real tics
    entertic = I_WaitForTic (oldentertics);
    realtics = entertic - oldentertics;
    oldentertics = entertic;

	DObject::BeginFrame ();

//	frameon++;

	realtics = realtics%BACKUPTICS;

    // run the realtics tics
    while (realtics--)
    {
        NetUpdate ();  // check for new console commands

		if (advancedemo)
			D_DoAdvanceDemo ();
        C_Ticker ();
        M_Ticker ();
        G_Ticker ();
        gametic++;
    }

	DObject::EndFrame ();
}

// [BC]
void CL_TryRunTics (void)
{
    int             entertic;
    static int      oldentertics;
    int             realtics;

    // get real tics
    entertic = I_WaitForTic (oldentertics);
    realtics = entertic - oldentertics;
    oldentertics = entertic;

	DObject::BeginFrame ();

    // run the realtics tics
    while (realtics--)
    {
        NetUpdate ();  // check for new console commands

        C_Ticker ();
        M_Ticker ();
        G_Ticker ();
        gametic++;
    }

	DObject::EndFrame ();
}


void Net_NewMakeTic (void)
{
	specials.NewMakeTic ();
}

//==========================================================================
//
// Dynamic buffer interface
//
//==========================================================================

FDynamicBuffer::FDynamicBuffer ()
{
	m_Data = NULL;
	m_Len = m_BufferLen = 0;
}

FDynamicBuffer::~FDynamicBuffer ()
{
	if (m_Data)
	{
		free (m_Data);
		m_Data = NULL;
	}
	m_Len = m_BufferLen = 0;
}

void FDynamicBuffer::SetData (const byte *data, int len)
{
	if (len > m_BufferLen)
	{
		m_BufferLen = (len + 255) & ~255;
		m_Data = (byte *)Realloc (m_Data, m_BufferLen);
	}
	if (data)
	{
		m_Len = len;
		memcpy (m_Data, data, len);
	}
	else
	{
		len = 0;
	}
}

byte *FDynamicBuffer::GetData (int *len)
{
	if (len)
		*len = m_Len;
	return m_Len ? m_Data : NULL;
}

// [BC]
bool CheckMe (const char *from, char *to)
{
	if (!strnicmp ("/me", from, 3))
	{
		from += 3;

		strcpy (to, from);
		return true;
	}

	return false;
}

//
// [BC] Net_Command
//
// Executes a network command right off the bat. bstr is a byte
// arg and cstr is a string arg. type is the type of special to activate,
// player is the playernum of the player who did the command
void Net_Command (int type, byte bstr, char *cstr, int player, BOOL force)
{
	if (!force)
	{
		if (netstate == NS_CLIENT)
		{
			char str[192];

			// [BC] Prevent chat flooding
			if (net_buffer.cursize > 300)
				return;

			if (cstr)
				strncpy (str, cstr, 192);

			MSG_WriteByte (&net_buffer, clc_netcmd);
			MSG_WriteShort (&net_buffer, (short)type);
			MSG_WriteByte (&net_buffer, bstr);
			MSG_WriteString(&net_buffer, str);
			MSG_WriteByte (&net_buffer, player);
			return;
		}

		if (netstate == NS_SERVER)
		{
			for (int i=0; i<MAXPLAYERS; i++)
			{
				client_t *cl = &clients[i];

				if (cl->cls < cs_spawned)
					continue;
		   
				MSG_WriteByte (&cl->netbuf, svc_netcmd);
				MSG_WriteShort (&cl->netbuf, (short)type);
				MSG_WriteByte (&cl->netbuf, bstr);
				MSG_WriteString(&cl->netbuf, cstr);
				MSG_WriteByte (&cl->netbuf, player);
			}
		}
	}

	switch (type)
	{
	case DEM_SAY:
	{
		// Said to everyone
		// [BC] /me support! ph33r!
		char mestr[256];

		if (oldschool.value)
		{
			char namestr[8];

			switch (player)
			{
			case 0:
				strcpy (namestr, "Green");
				break;
			case 1:
				strcpy (namestr, "Indigo");
				break;
			case 2:
				strcpy (namestr, "Brown");
				break;
			case 3:
				strcpy (namestr, "Red");
				break;
			}

			Printf (PRINT_CHAT, "%s: %s\n", namestr, cstr);
		}

		else if (CheckMe (cstr, mestr))
			Printf (PRINT_HIGH, "* %s%s\n", players[player].userinfo.netname, mestr);
		else
			Printf (PRINT_CHAT, "%s: %s\n", players[player].userinfo.netname, cstr);
		S_Sound (CHAN_VOICE, gameinfo.chatSound, 1, ATTN_NONE);

		// Handle bot responses
		for (int i=0; i<MAXPLAYERS; i++)
		{
			if (!playeringame[i] || !players[i].skullb)
				continue;

			for (int j=0; j<NUMBOTSTRINGS; j++)
			{
				if (players[i].skullb->info.chatinfo == botstrings[j].bot)
				{
					if (!botstrings[j].string)
						continue;

					if (!strnicmp(cstr, botstrings[j].string, strlen(botstrings[j].string)))
					{
						players[i].skullb->t_chat = players[i].skullb->info.chattime;
						players[i].skullb->chatline = botstrings[j].line;
					}
				}
			}
		}

		if (players[player].icon)
		{
			players[player].icon->Destroy ();
			players[player].icon = NULL;
		}
	}
		break;

	case DEM_MUSICCHANGE:
		S_ChangeMusic (cstr, true);
		break;

	case DEM_PRINT:
		Printf (PRINT_HIGH, cstr);
		break;

	case DEM_CENTERPRINT:
		C_MidPrint (cstr);
		break;

	case DEM_UINFCHANGED:
		D_ReadUserInfoStrings (player, cstr, true);
		break;

	case DEM_SINFCHANGED:
		D_DoServerInfoChange (cstr);
		break;

	case DEM_GIVECHEAT:
		cht_Give (&players[player], cstr);
		break;

	case DEM_GENERICCHEAT:
		cht_DoCheat (&players[player], bstr);
		break;

	case DEM_CHANGEMAP:
		// Change to another map without disconnecting other players
		strncpy (level.nextmap, cstr, 8);
		level.flags |= LEVEL_CHANGEMAPCHEAT;
		G_ExitLevel (0);
		break;

	case DEM_ADDBOT:
//		B_AddBot (bstr, cstr, cstr);
		break;

	case DEM_KILLBOTS:
		B_RemoveBots ();
		if (level.time > 3*TICRATE)
			Printf (PRINT_HIGH, "Removed all bots\n");
		break;

	default:
		I_Error ("Unknown net command: %d", type);
		break;
	}
}
